import json
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_instance_profile_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "InstanceProfileName": "name",
            "InstanceProfileId": "id",
            "Path": "path",
            "Tags": "tags",
        }
    )

    # Name is the unique identifier for instance_profile so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("InstanceProfileName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    translated_resource["roles"] = []
    for role in raw_resource.get("Roles"):
        translated_resource["roles"].append({"RoleName": role.get("RoleName")})

    return translated_resource


async def convert_raw_policy_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "PolicyName": "name",
            "Arn": "arn",
            "PolicyId": "id",
            "Path": "path",
            "Description": "description",
            "Tags": "tags",
        }
    )

    policy_arn = raw_resource.get("Arn")
    # Arn is the unique identifier for policy so it is set as resource_id
    translated_resource = {"resource_id": policy_arn}

    default_version_id = raw_resource.get("DefaultVersionId")
    ret_get = await hub.exec.boto3.client.iam.get_policy(ctx, PolicyArn=policy_arn)
    if not ret_get["result"]:
        hub.log.warning(
            f"Failed on fetching policy with arn {policy_arn} with error {ret_get['comment']}."
        )
        return None

    ret_get_version = await hub.exec.boto3.client.iam.get_policy_version(
        ctx, PolicyArn=policy_arn, VersionId=default_version_id
    )
    if not ret_get_version["result"]:
        hub.log.warning(
            f"Failed on fetching policy document with arn {policy_arn} and version {default_version_id} "
            f"with error {ret_get_version['comment']} Describe will skip this policy and continue."
        )
        return None

    document = json.dumps(ret_get_version["ret"]["PolicyVersion"].get("Document"))
    translated_resource["policy_document"] = document
    translated_resource["default_version_id"] = default_version_id

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    return translated_resource


def convert_raw_role_to_present(hub, raw_resource: Dict[str, Any]) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "RoleName": "name",
            "Arn": "arn",
            "RoleId": "id",
            "Path": "path",
            "Description": "description",
            "MaxSessionDuration": "max_session_duration",
            "Tags": "tags",
        }
    )

    # RoleName is the unique identifier for policy so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("RoleName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("PermissionsBoundary"):
        translated_resource["permissions_boundary"] = raw_resource.get(
            "PermissionsBoundary"
        ).get("PermissionsBoundaryArn")

    if raw_resource.get("AssumeRolePolicyDocument"):
        translated_resource["assume_role_policy_document"] = json.dumps(
            raw_resource.get("AssumeRolePolicyDocument")
        )

    return translated_resource


def convert_raw_role_policy_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    # Policy name is the 'name' and 'resource_id' of the role_policy
    resource_parameters = OrderedDict(
        {
            "RoleName": "role_name",
            "PolicyName": "name",
        }
    )

    # Use policy name as resource_id
    translated_resource = {"resource_id": raw_resource.get("PolicyName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("PolicyDocument"):
        translated_resource["policy_document"] = json.dumps(
            raw_resource.get("PolicyDocument")
        )

    return translated_resource


def convert_raw_role_policy_attachment_to_present(
    hub, role_name: str, policy_arn: str
) -> Dict[str, Any]:

    translated_resource = {"role_name": role_name, "policy_arn": policy_arn}

    return translated_resource



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_nat_gateway
hub.exec.boto3.client.ec2.delete_nat_gateway
hub.exec.boto3.client.ec2.describe_nat_gateways
"""
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "present": {
        "require": ["aws.ec2.subnet.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_id: str,
    connectivity_type: str,
    client_token: str = None,
    allocation_id: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a NAT gateway in the specified subnet. This action creates a network interface in the specified subnet
    with a private IP address from the IP address range of the subnet. You can create either a public NAT gateway or
    a private NAT gateway. With a public NAT gateway, internet-bound traffic from a private subnet can be routed to
    the NAT gateway, so that instances in a private subnet can connect to the internet. With a private NAT gateway,
    private communication is routed across VPCs and on-premises networks through a transit gateway or virtual
    private gateway. Common use cases include running large workloads behind a small pool of allowlisted IPv4
    addresses, preserving private IPv4 addresses, and communicating between overlapping networks. For more
    information, see NAT gateways in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): The ID of the NAT gateway.
        subnet_id(Text): The subnet in which to create the NAT gateway.
        client_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Constraint: Maximum 64 ASCII characters. This field is autopopulated if not provided.
        connectivity_type(Text, optional): Indicates whether the NAT gateway supports public or private connectivity. The default is public connectivity.
        allocation_id(Text, optional): [Public NAT gateway only]. The allocation ID of the Elastic IP address that's associated with the NAT gateway.
        tags(List, optional): The tags to assign to the subnet. Defaults to None.
            * Key (string) --  The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [nat_gateway-id]:
          aws.ec2.nat_gateway.present:
          - subnet_id: 'string'
          - client_token: 'string'
          - connectivity_type: 'string'
          - allocation_id: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nat-0e7969ddf4c38831f:
              aws.ec2.nat_gateway.present:
                - subnet_id: subnet-026542dd856a97e15
                - client-token: nat-0e7969ddf4c38831f
                - connectivity_type: private
                - tags:
                  - Key: Name
                    Value: test-nat-gateway
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    nat_gateway_id = None
    before = await hub.exec.boto3.client.ec2.describe_nat_gateways(
        ctx, NatGatewayIds=[name]
    )
    if before["result"] and before["ret"].get("NatGateways"):
        # check if tags need to be updated
        if tags is not None:
            nat_gateway_id = name
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=name,
                old_tags=before["ret"]["NatGateways"][0]["Tags"],
                new_tags=tags,
            )
            result["comment"] = f"'{name}' already present. Updating tags"
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_nat_gateway(
                ctx,
                AllocationId=allocation_id,
                ClientToken=client_token,
                SubnetId=subnet_id,
                TagSpecifications=[{"ResourceType": "natgateway", "Tags": tags}]
                if tags is not None
                else None,
                ConnectivityType=connectivity_type,
            )
            result["result"] = ret["ret"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            nat_gateway_id = ret["ret"]["NatGateway"]["NatGatewayId"]
            result["comment"] = f"Created '{nat_gateway_id}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    after = None
    if nat_gateway_id is not None:
        after = await hub.exec.boto3.client.ec2.describe_nat_gateways(
            ctx, NatGatewayIds=[nat_gateway_id]
        )
    if before["result"] and before["ret"].get("NatGateways"):
        result["old_state"] = before["ret"]["NatGateways"][0]
    if after and after["result"] and after["ret"].get("NatGateways"):
        result["new_state"] = after["ret"]["NatGateways"][0]
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified NAT gateway. Deleting a public NAT gateway disassociates its Elastic IP address, but does
    not release the address from your account. Deleting a NAT gateway does not delete any NAT gateway routes in your
    route tables.

    Args:
        name(Text): The ID of the NAT gateway.

    Request Syntax:
        [nat_gateway-id]:
          aws.ec2.nat_gateway.absent:
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nat-0e7969ddf4c38831f:
                aws.ec2.nat_gateway.absent:
                  - name: nat-0e7969ddf4c38831f
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    nat_gateway_id_list = [name]
    before = await hub.exec.boto3.client.ec2.describe_nat_gateways(
        ctx, NatGatewayIds=nat_gateway_id_list
    )
    if not before["result"]:
        result["comment"] = f"'{name}' already absent"
    else:
        try:
            if before["ret"]["NatGateways"][0]["State"] == "deleted":
                result["comment"] = f"'{name}' is already in deleted state."
            else:
                ret = await hub.exec.boto3.client.ec2.delete_nat_gateway(
                    ctx, NatGatewayId=name
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    if before["result"] and before["ret"].get("NatGateways"):
        result["old_state"] = before["ret"]["NatGateways"][0]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your NAT gateways.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.nat_gateway
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_nat_gateways(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe nat_gateway {ret['comment']}")
        return result

    describe_parameters = OrderedDict(
        {
            "SubnetId": "subnet_id",
            "ConnectivityType": "connectivity_type",
            "Tags": "tags",
        }
    )

    for nat_gateway in ret["ret"]["NatGateways"]:
        nat_gateway_id = nat_gateway.get("NatGatewayId")
        resource_translated = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if (
                nat_gateway.get(parameter_old_key)
                and nat_gateway.get(parameter_old_key) is not None
            ):
                resource_translated.append(
                    {parameter_new_key: nat_gateway.get(parameter_old_key)}
                )
        if nat_gateway.get("NatGatewayAddresses"):
            for nat_gateway_address in nat_gateway.get("NatGatewayAddresses"):
                if "AllocationId" in nat_gateway_address:
                    resource_translated.append(
                        {"allocation_id": nat_gateway_address.get("AllocationId")}
                    )
                    break
        result[nat_gateway_id] = {"aws.ec2.nat_gateway.present": resource_translated}
    return result



import copy


async def update_rules_in_sg(hub, ctx, created_sg_id, sg_rules, resource):
    """
    Modify, Add or Delete rules in the security group.

    Args:
         hub:
         ctx:
         created_sg_id: Security group ID to which the rules are attached.
         sg_rules: List of new security group rules
         resource: Security group resource to perform operation on
    """
    result = dict(comment="", result=True, ret=None)
    new_rules = []
    modify_rules = []
    result["comment"] = f"'{created_sg_id}' already exists"
    existing_sg_rules_result = (
        await hub.exec.boto3.client.ec2.describe_security_group_rules(
            ctx,
            Filters=[{"Name": "group-id", "Values": [created_sg_id]}],
        )
    )
    if existing_sg_rules_result["result"]:
        existing_sg_rules = existing_sg_rules_result["ret"].get("SecurityGroupRules")
        delete_egress_sg_rule_ids = []
        delete_ingress_sg_rule_ids = []
        for rule in existing_sg_rules:
            if rule["IsEgress"]:
                delete_egress_sg_rule_ids.append(rule["SecurityGroupRuleId"])
            else:
                delete_ingress_sg_rule_ids.append(rule["SecurityGroupRuleId"])
        if existing_sg_rules != sg_rules:
            for security_grp_rule in sg_rules:
                if "SecurityGroupRuleId" in security_grp_rule:
                    if security_grp_rule["IsEgress"]:
                        delete_egress_sg_rule_ids.remove(
                            security_grp_rule["SecurityGroupRuleId"]
                        )
                    else:
                        delete_ingress_sg_rule_ids.remove(
                            security_grp_rule["SecurityGroupRuleId"]
                        )
                    if security_grp_rule not in existing_sg_rules:
                        modify_rules.append(security_grp_rule)
                else:
                    new_rules.append(security_grp_rule)
        if new_rules:
            new_rules_dict = hub.tool.aws.security_group_utils.convert_sg_rules_to_ingress_and_egress(
                new_rules
            )
            existing_rules_dict = hub.tool.aws.security_group_utils.convert_sg_rules_to_ingress_and_egress(
                existing_sg_rules
            )
            existing_ingress_rules = existing_rules_dict["ingress"]
            converted_ingress_existed_rules = (
                hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    existing_ingress_rules
                )
            )
            for rule in converted_ingress_existed_rules:
                rule.pop("SecurityGroupRuleId", None)
            if new_rules_dict["ingress"]:
                new_ingress_rules = new_rules_dict["ingress"]
                final_ingress_rules = [
                    item
                    for item in new_ingress_rules
                    if item not in converted_ingress_existed_rules
                ]
                new_ingress_rules = hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    final_ingress_rules
                )
                ingress_result = (
                    await hub.exec.boto3.client.ec2.authorize_security_group_ingress(
                        ctx,
                        GroupId=created_sg_id,
                        IpPermissions=new_ingress_rules,
                    )
                )

                if not ingress_result[
                    "result"
                ] and "InvalidPermission.Duplicate" not in str(
                    ingress_result["comment"]
                ):
                    result["comment"] = ingress_result["comment"]
                    result["result"] = False
            if new_rules_dict["egress"]:
                new_egress_rules = hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    new_rules_dict["egress"]
                )
                existing_egress_rules = existing_rules_dict["egress"]
                converted_egress_existed_rules = hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    existing_egress_rules
                )
                for rule in converted_egress_existed_rules:
                    rule.pop("SecurityGroupRuleId", None)
                final_egress_rules = [
                    item
                    for item in new_egress_rules
                    if item not in converted_egress_existed_rules
                ]
                egress_result = (
                    await hub.exec.boto3.client.ec2.authorize_security_group_egress(
                        ctx,
                        GroupId=created_sg_id,
                        IpPermissions=final_egress_rules,
                    )
                )
                if not egress_result[
                    "result"
                ] and "InvalidPermission.Duplicate" not in str(
                    egress_result["comment"]
                ):
                    result["comment"] = egress_result["comment"]
                    result["result"] = False
        if modify_rules:
            new_modified_rules = []
            existing_sg_rules_copy = copy.deepcopy(existing_sg_rules)
            for rule in existing_sg_rules_copy:
                rule.pop("SecurityGroupRuleId", None)
            final_modified_rules = [
                item for item in modify_rules if item not in existing_sg_rules_copy
            ]
            for modify_rule in final_modified_rules:
                new_modified_rule = dict(
                    SecurityGroupRuleId=None, SecurityGroupRule=None
                )
                new_modified_rule["SecurityGroupRuleId"] = modify_rule[
                    "SecurityGroupRuleId"
                ]
                updated_rule = {}
                if modify_rule.get("IpProtocol"):
                    updated_rule["IpProtocol"] = modify_rule.get("IpProtocol")
                if modify_rule.get("FromPort"):
                    updated_rule["FromPort"] = modify_rule.get("FromPort")
                if modify_rule.get("ToPort"):
                    updated_rule["ToPort"] = modify_rule.get("ToPort")
                if modify_rule.get("CidrIpv4"):
                    updated_rule["CidrIpv4"] = modify_rule.get("CidrIpv4")
                if modify_rule.get("CidrIpv6"):
                    updated_rule["CidrIpv6"] = modify_rule.get("CidrIpv6")
                if modify_rule.get("PrefixListId"):
                    updated_rule["PrefixListId"] = modify_rule.get("PrefixListId")
                if modify_rule.get("ReferencedGroupId"):
                    updated_rule["ReferencedGroupId"] = modify_rule.get(
                        "ReferencedGroupId"
                    )
                if modify_rule.get("Description"):
                    updated_rule["Description"] = modify_rule.get("Description")
                new_modified_rule["SecurityGroupRule"] = updated_rule
                new_modified_rules.append(new_modified_rule)
            modify_rules_result = (
                await hub.exec.boto3.client.ec2.modify_security_group_rules(
                    ctx,
                    GroupId=created_sg_id,
                    SecurityGroupRules=new_modified_rules,
                )
            )
            if not modify_rules_result[
                "result"
            ] and "InvalidPermission.Duplicate" not in str(
                modify_rules_result["comment"]
            ):
                result["comment"] = modify_rules_result["comment"]
                result["result"] = False
        if delete_ingress_sg_rule_ids:
            delete_ingress_result = resource.revoke_ingress(
                SecurityGroupRuleIds=delete_ingress_sg_rule_ids
            )
            if not delete_ingress_result["Return"]:
                result["comment"] = delete_ingress_result["ResponseMetadata"]
                result["result"] = False
        if delete_egress_sg_rule_ids:
            delete_egress_result = resource.revoke_egress(
                SecurityGroupRuleIds=delete_egress_sg_rule_ids
            )
            if not delete_egress_result["Return"]:
                result["comment"] = delete_egress_result["ResponseMetadata"]
                result["result"] = False
    else:
        result["comment"] = existing_sg_rules_result["comment"]
        result["result"] = False
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elasticache.create_cache_subnet_group
hub.exec.boto3.client.elasticache.delete_cache_subnet_group
hub.exec.boto3.client.elasticache.describe_cache_subnet_groups
hub.exec.boto3.client.elasticache.modify_cache_subnet_group
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_ids: List,
    cache_subnet_group_description: str,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new cache subnet group. Use cache subnet group parameter only when you are creating a cluster in an Amazon Virtual
    Private Cloud (Amazon VPC).

    Args:
        name(Text): A name, ID to identify the resource.
        cache_subnet_group_description(Text): Description of the cache subnet group
        subnet_ids(list): A list of VPC subnet IDs for the cache subnet group.
        tags(List, optional): A list of tags to be added to this resource.
            Key (string) - The key for the tag. May not be null.
            Value (string) - The tag's value. May be null.

    Request Syntax:
        [cache_subnet_group_name]:
          aws.elasticache.cache_subnet_group.present:
          - name: 'string'
          - cache_subnet_group_description: 'string'
          - subnet_ids: 'List'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cache_subnet_group_name:
              aws.elasticache.cache_subnet_group.present:
                - cache_subnet_group_description: "testing demo"
                - subnet_ids:
                    - subnet-239e4b44
                - tags:
                    - Key: Name
                      Value: test-elasticache
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)
    update_new_state = True
    before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
        ctx, CacheSubnetGroupName=name
    )
    try:
        if before["result"]:
            result["old_state"] = before["ret"]["CacheSubnetGroups"][0]
            result["comment"] = f"'{name}' already present."
            update_ret = (
                await hub.exec.boto3.client.elasticache.modify_cache_subnet_group(
                    ctx=ctx,
                    CacheSubnetGroupName=name,
                    CacheSubnetGroupDescription=cache_subnet_group_description,
                    SubnetIds=subnet_ids,
                )
            )
            if update_ret["result"]:
                result["comment"] = f"'Updated {name}'"
            else:
                update_new_state = False
                result["result"] = False
            resource_arn = before["ret"]["CacheSubnetGroups"][0]["ARN"]
            old_tags = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
                ctx, ResourceName=resource_arn
            )
            if tags is not None and old_tags["ret"]["TagList"] != tags:
                # Update tags
                result["comment"] = f"'{name}' already present. Updating tags"
                update_new_state = True
                update_tag = await hub.exec.aws.elasticache.tag.update_tags(
                    ctx=ctx,
                    resource_arn=resource_arn,
                    old_tags=old_tags["ret"]["TagList"],
                    new_tags=tags,
                )
                if not update_tag["result"]:
                    result["comment"] = update_tag["comment"]
                    update_new_state = False
                    result["result"] = False
        else:
            ret = await hub.exec.boto3.client.elasticache.create_cache_subnet_group(
                ctx,
                CacheSubnetGroupName=name,
                CacheSubnetGroupDescription=cache_subnet_group_description,
                SubnetIds=subnet_ids,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = f"{e.__class__.__name__}: {e}"
    try:
        if update_new_state:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
                    ctx, CacheSubnetGroupName=name
                )
            )
            result["new_state"] = after["ret"]["CacheSubnetGroups"][0]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a cache subnet group.  You cannot delete a default cache subnet group or one that is associated with any
    clusters.

    Args:
        name(Text): A name, ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.elasticache.cache_subnet_group.absent:
                - name: value
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)

    before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
        ctx, CacheSubnetGroupName=name
    )
    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.elasticache.cache_subnet_group {name}"
        return result
    else:
        result["old_state"] = before["ret"]["CacheSubnetGroups"][0]
        try:
            ret = await hub.exec.boto3.client.elasticache.delete_cache_subnet_group(
                ctx, CacheSubnetGroupName=name
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of cache subnet group descriptions. If a subnet group name is specified, the list contains only
    the description of that group. This is applicable only when you have ElastiCache in VPC setup. All ElastiCache
    clusters now launch in VPC by default.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.elasticache.cache_subnet_group
    """

    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe cache_subnet_group {ret['comment']}")
        return {}

    describe_parameters = OrderedDict(
        {
            "CacheSubnetGroupDescription": "cache_subnet_group_description",
        }
    )
    for subnet_group in ret["ret"]["CacheSubnetGroups"]:
        CacheSubnetGroupName = subnet_group.get("CacheSubnetGroupName")

        resource_translated = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if subnet_group.get(parameter_old_key) is not None:
                resource_translated.append(
                    {parameter_new_key: subnet_group.get(parameter_old_key)}
                )
        if subnet_group.get("Subnets"):
            subnet_ids_list = []
            for nat_gateway_address in subnet_group.get("Subnets"):
                if "SubnetIdentifier" in nat_gateway_address:
                    subnet_ids_list.append(nat_gateway_address.get("SubnetIdentifier"))
            resource_translated.append({"subnet_ids": subnet_ids_list})
        if subnet_group.get("ARN"):
            tags = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
                ctx, ResourceName=subnet_group.get("ARN")
            )
            if tags["ret"]["TagList"]:
                resource_translated.append({"tags": tags["ret"]["TagList"]})
        result[CacheSubnetGroupName] = {
            "aws.elasticache.cache_subnet_group.present": resource_translated
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS EC2 to present input format.
"""


def convert_raw_vpc_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("VpcId")
    resource_parameters = OrderedDict(
        {"InstanceTenancy": "instance_tenancy", "Tags": "tags"}
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    # The following code block is to make sure to only output the associated/associating cidr blocks and
    # dis-regard the disassociated cidr blocks.
    if raw_resource.get("CidrBlockAssociationSet"):
        ipv4_cidr_block_association_set = []
        for cidr_block in raw_resource.get("CidrBlockAssociationSet"):
            if cidr_block.get("CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                ipv4_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "cidr_block_association_set"
        ] = ipv4_cidr_block_association_set
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = []
        for cidr_block in raw_resource.get("Ipv6CidrBlockAssociationSet"):
            if cidr_block.get("Ipv6CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                if "NetworkBorderGroup" in cidr_block:
                    # Translate describe output to the correct present input format
                    ipv6_cidr_block_network_border_group = cidr_block.pop(
                        "NetworkBorderGroup"
                    )
                    cidr_block[
                        "Ipv6CidrBlockNetworkBorderGroup"
                    ] = ipv6_cidr_block_network_border_group
                ipv6_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "ipv6_cidr_block_association_set"
        ] = ipv6_cidr_block_association_set
    return resource_translated


def convert_raw_subnet_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SubnetId")
    resource_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "CidrBlock": "cidr_block",
            "AvailabilityZone": "availability_zone",
            "OutpostArn": "outpost_arn",
            "Tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if (not raw_resource.get("AvailabilityZone")) and raw_resource.get(
        "AvailabilityZoneId"
    ):
        # Only populate availability_zone_id field when availability_zone doesn't exist
        resource_translated["availability_zone_id"] = raw_resource.get(
            "AvailabilityZoneId"
        )
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = (
            hub.tool.aws.network_utils.get_associated_ipv6_cidr_blocks(
                raw_resource.get("Ipv6CidrBlockAssociationSet")
            )
        )
        # We should only output the associated ipv6 cidr block, and theoretically there should only be one,
        # since AWS only supports one ipv6 cidr block association on a subnet
        if ipv6_cidr_block_association_set:
            resource_translated["ipv6_cidr_block"] = ipv6_cidr_block_association_set[
                0
            ].get("Ipv6CidrBlock")
    return resource_translated



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

from dict_tools import differ

SERVICE = "iam"
RESOURCE = "User"

__contracts__ = ["resource"]


async def absent(hub, ctx, name: str, **kwargs) -> Dict[str, Any]:
    ret = {
        "comment": "",
        "old_state": None,
        "new_state": None,
        "name": name,
        "result": True,
    }
    resource = hub.tool.boto3.resource.create(ctx, SERVICE, RESOURCE, name=name)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        ret["comment"] = f"{RESOURCE} '{name}' already absent"
    else:
        try:
            before.pop("ResponseMetadata", None)
            ret["old_state"] = before
            # Attempt to delete the User
            await hub.tool.boto3.resource.exec(resource, "delete", **kwargs)
            ret["comment"] = f"Delete {RESOURCE} '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            ret["comment"] = f"{e.__class__.__name__: {e}}"
            ret["result"] = False
    return ret


async def present(hub, ctx, name: str, **kwargs) -> Dict[str, Any]:
    ret = {
        "comment": "",
        "old_state": None,
        "new_state": None,
        "name": name,
        "result": True,
    }
    resource = hub.tool.boto3.resource.create(ctx, SERVICE, RESOURCE, name=name)
    before = await hub.tool.boto3.resource.describe(resource)

    if before:
        ret["comment"] = f"{RESOURCE} '{name}' already exists"
        before.pop("ResponseMetadata", None)
        ret["old_state"] = before
        ret["new_state"] = copy.deepcopy(before)
    else:
        try:
            # Attempt to create the User
            await hub.tool.boto3.resource.exec(resource, "create", **kwargs)
            ret["comment"] = f"Created {RESOURCE}, '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            # User already exists
            ret["comment"] = f"{e.__class__.__name__: {e}}"

        await hub.tool.boto3.client.wait(ctx, SERVICE, "user_exists", UserName=name)

        after = await hub.tool.boto3.resource.describe(resource)
        if after:
            ret["result"] = True
            after.pop("ResponseMetadata", None)
            ret["new_state"] = after
        else:
            ret["result"] = False
    return ret


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.iam.list_users(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe user {ret['comment']}")
        return {}

    # Arn is not used for present but required for arg binding
    describe_parameters = OrderedDict(
        {
            "Arn": "arn",
            "CreateDate": "create_date",
            "PasswordLastUsed": "password_last_used",
            "Path": "path",
            "PermissionsBoundary": "permissions_boundary",
            "Tags": "tags",
            "UserId": "user_id",
            "UserName": "user_name",
        }
    )

    for user in ret["ret"]["Users"]:
        resource_name = user.get("UserName")
        # This is required to get tags for each user
        boto2_resource = hub.tool.boto3.resource.create(
            ctx, "iam", "User", resource_name
        )
        resource = await hub.tool.boto3.resource.describe(boto2_resource)
        resource_translated = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if resource.get(parameter_old_key) is not None:
                resource_translated.append(
                    {parameter_new_key: resource.get(parameter_old_key)}
                )
        if resource.get("PermissionsBoundary"):
            resource_translated.append(
                {
                    "permissions_boundary": resource.get("PermissionsBoundary").get(
                        "PermissionsBoundaryArn"
                    )
                }
            )
        result[resource_name] = {"aws.iam.user.present": resource_translated}
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.attach_policy
hub.exec.boto3.client.organizations.create_policy
hub.exec.boto3.client.organizations.delete_policy
hub.exec.boto3.client.organizations.describe_policy
hub.exec.boto3.client.organizations.detach_policy
hub.exec.boto3.client.organizations.list_policies
hub.exec.boto3.client.organizations.update_policy
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization.present",
            "aws.organizations.organization_unit.present",
            "aws.organizations.account.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    policy_name: str = None,
    description: str = None,
    policy_type: str = None,
    content: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a policy of a specified type that you can attach to a root, an organizational unit (OU), or an
    individual AWS account. For more information about policies and their use, see Managing Organization Policies.
    If the request includes tags, then the requester must have the organizations:TagResource permission. This
    operation can be called only from the organization's management account.

    Args:
        name(Text): An ID to identify the resource.
        policy_name: The friendly name to assign to the policy.
        description: An optional description to assign to the policy.
        policy_type : The type of policy to create. Only supported type currently is SERVICE_CONTROL_POLICY
        content : The policy text content to add to the new policy. The text that you supply must adhere to
                    the rules of the policy type you specify in the Type parameter
        tags(Optional): A list of tags that you want to attach to the newly created policy

    Request Syntax:
        [policy-id]:
          aws.organizations.policy.present:
          - policy_name: 'string'
          - description: 'string'
          - policy_type: 'string'
          - content: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            p-id:
              aws.organizations.policy.present:
                - policy_name: AllowAllS3Actions
                - description: Enables admins of attached accounts to delegate all S3 permissions
                - policy_type : SERVICE_CONTROL_POLICY
                - content : '{\"Version\":\"2012-10-17\",\"Statement\":{\"Effect\":\"Allow\",\"Action\":\"s3:*\"}}'
                - tags:
                  - Key: org
                    Value: idem-aws
    """

    result = dict(comment="", name=name, result=True, old_state=None, new_state=None)

    before = await hub.exec.boto3.client.organizations.describe_policy(
        ctx, PolicyId=name
    )

    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.organizations.policy {name}."
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.oranizations.policy {name}."
            result["result"] = True
        return result

    change_dict = dict(Name=None, Description=None, Content=None)

    update_tag = False
    policy_id = None
    if before:
        # Policy exists , update
        result["comment"] = f"aws.organizations.policy '{name}' exists."
        before["ret"].pop("ResponseMetadata", None)
        result["old_state"] = before["ret"]["Policy"]["PolicySummary"]
        result["old_state"]["Content"] = before["ret"]["Policy"]["Content"]

        policy_id = before["ret"]["Policy"]["PolicySummary"].get("Id")

        try:

            if before["ret"]["Policy"]["PolicySummary"].get("Name") != policy_name:
                change_dict["Name"] = policy_name

            if (
                before["ret"]["Policy"]["PolicySummary"].get("Description")
                != description
            ):
                change_dict["Description"] = description

            if before["ret"]["Policy"]["Content"] != content:
                change_dict["Content"] = content

            if any(value is not None for value in change_dict.values()):

                update_policy_ret = (
                    await hub.exec.boto3.client.organizations.update_policy(
                        ctx,
                        PolicyId=policy_id,
                        Name=change_dict.get("Name"),
                        Description=change_dict.get("Description"),
                        Content=change_dict.get("Content"),
                    )
                )
                result["result"] = update_policy_ret["result"]
                if not update_policy_ret:
                    result["comment"] = update_policy_ret["comment"]
                    return result

                after = update_policy_ret

                after["ret"].pop("ResponseMetadata", None)
                result["new_state"] = after["ret"]["Policy"]["PolicySummary"]
                result["new_state"]["Content"] = after["ret"]["Policy"]["Content"]

                result["comment"] = f"Updated aws.organizations.policy '{name}'"

            else:
                result["new_state"] = copy.deepcopy(result["old_state"])

            old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
                ctx, ResourceId=policy_id
            )

            if old_tags:
                before["ret"]["Policy"]["Tags"] = old_tags["ret"]["Tags"]
                result["old_state"]["Tags"] = old_tags["ret"]["Tags"]

                # If no update in tags , then new_state will contain same tags as old
                if tags is None:
                    result["new_state"]["Tags"] = result["old_state"]["Tags"]

            if tags is not None and old_tags["result"]:

                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, policy_id, old_tags["ret"].get("Tags", []), tags
                    )
                )
                if not update_tags_ret["result"]:
                    result[
                        "comment"
                    ] = f" {result['comment']} , {str(update_tags_ret['comment'])}"
                    result["result"] = False
                    return result

                if update_tags_ret["ret"] is not None:
                    update_tag = True

                    old_comment = result["comment"]
                    result[
                        "comment"
                    ] = f"'{old_comment}'. Updated tags on aws.organizations.account '{name}'."

            result["result"] = True
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = f"{e.__class__.__name__}: {e}"

    else:

        # Policy does not exist , create
        try:
            create_policy_ret = await hub.exec.boto3.client.organizations.create_policy(
                ctx,
                Name=policy_name,
                Description=description,
                Type=policy_type,
                Content=content,
                Tags=tags,
            )
            result["result"] = create_policy_ret["result"]
            if not result["result"]:
                result["comment"] = create_policy_ret["comment"]
                return result

            after = create_policy_ret

            after["ret"].pop("ResponseMetadata", None)
            result["new_state"] = after["ret"]["Policy"]["PolicySummary"]
            result["new_state"]["Content"] = after["ret"]["Policy"]["Content"]

            policy_id = result["new_state"]["Id"]
            if tags is not None:
                list_tags = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=policy_id
                    )
                )
                if list_tags:
                    result["new_state"]["Tags"] = list_tags["ret"]["Tags"]

            result["comment"] = f"Created aws.organizations.policy '{name}'."
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    if before and update_tag:
        try:
            updated_tags = (
                await hub.exec.boto3.client.organizations.list_tags_for_resource(
                    ctx, ResourceId=policy_id
                )
            )

            if updated_tags:
                result["new_state"]["Tags"] = updated_tags["ret"]["Tags"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified policy from your organization. Before you perform this operation, you must first detach
    the policy from all organizational units (OUs), roots, and accounts. This operation can be called only from the
    organization's management account.

    Args:
        name(Text): An Identifier to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
            p-iauqv2gg:
                aws.organizations.policy.absent:
                - name: p-iauqv2gg
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_policy(
        ctx, PolicyId=name
    )

    policy_id = name

    if not before:
        result["comment"] = f"aws.organizations.policy {name} already absent"
        result["result"] = True
    else:
        try:

            delete_ret = await hub.exec.boto3.client.organizations.delete_policy(
                ctx, PolicyId=policy_id
            )

            result["result"] = delete_ret["result"]

            if not result["result"]:
                result["comment"] = delete_ret["comment"]
                return result

            before["ret"].pop("ResponseMetadata", None)

            result["old_state"] = before["ret"]["Policy"]["PolicySummary"]
            result["old_state"]["Content"] = before["ret"]["Policy"]["Content"]

            result["comment"] = f"aws.organizations.policy {name} deleted."
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Retrieves information about a policy.

    This operation can be called only from the organization's management account
    or by a member account that is a delegated administrator for an AWS service.

    Currently IDEM AWS supports only SERVICE_CONTROL_POLICY.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.policy
    """

    result = {}

    list_all_policies = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="SERVICE_CONTROL_POLICY"
    )

    if not list_all_policies:
        hub.log.debug(f"Could not describe policy {list_all_policies['comment']}")
        return result

    describe_parameters = OrderedDict(
        {
            "Name": "policy_name",
            "Description": "description",
            "Type": "policy_type",
            "Content": "content",
            "Tags": "tags",
        }
    )

    for policy in list_all_policies["ret"]["Policies"]:
        policy_id = policy["Id"]
        policy_ret = await hub.exec.boto3.client.organizations.describe_policy(
            ctx, PolicyId=policy_id
        )

        if not policy_ret and not policy_ret["ret"]:
            hub.log.debug(f"Could not describe '{policy_id}'")
            continue

        translated_resource = []

        policy_desc = policy_ret["ret"]["Policy"]

        for camel_case_key, snake_case_key in describe_parameters.items():
            if policy_desc["PolicySummary"].get(camel_case_key) is not None:
                translated_resource.append(
                    {snake_case_key: policy_desc["PolicySummary"].get(camel_case_key)}
                )
            elif policy_desc.get(camel_case_key) is not None:
                translated_resource.append(
                    {snake_case_key: policy_desc.get(camel_case_key)}
                )

        tags_ret = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=policy_id
        )
        if tags_ret:
            translated_resource.append({"tags": tags_ret["ret"]["Tags"]})

        result[policy["Id"]] = {"aws.organizations.policy.present": translated_resource}

    return result



def convert_sg_rules_to_ingress_and_egress(hub, sg_rules: list):
    """
    Convert list of security group rules into egress and ingress list

     Args:
        hub: required for functions in hub
        sg_rules: list of security group rules

    Returns: {"ingress": List, "egress": List}
    """
    result = {"ingress": None, "egress": None}
    ingress = []
    egress = []
    for sg_rule in sg_rules:
        if sg_rule.get("CidrIpv4"):
            cidr_ip = sg_rule["CidrIpv4"]
            sg_rule["IpRanges"] = [{"CidrIp": cidr_ip}]
        elif sg_rule.get("CidrIpv6"):
            cidr_ip = sg_rule["CidrIpv6"]
            sg_rule["Ipv6Ranges"] = [{"CidrIpv6": cidr_ip}]
        if bool(sg_rule["IsEgress"]):
            egress.append(sg_rule)
        else:
            ingress.append(sg_rule)
    result["ingress"] = ingress
    result["egress"] = egress
    return result


def remove_unwanted_attributes_from_rule(hub, rules: list):
    """
    Remove unwanted params from list of rules

     Args:
        hub: required for functions in hub
        rules: list of security group rules

    Returns: A list of security group rules
    """
    for rule in rules:
        rule.pop("IsEgress", None)
        rule.pop("Tags", None)
        rule.pop("CidrIpv4", None)
        rule.pop("CidrIpv6", None)
        rule.pop("GroupId", None)
        rule.pop("GroupOwnerId", None)
    return rules



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.attach_internet_gateway
hub.exec.boto3.client.ec2.create_internet_gateway
hub.exec.boto3.client.ec2.delete_internet_gateway
hub.exec.boto3.client.ec2.describe_internet_gateways
hub.exec.boto3.client.ec2.detach_internet_gateway
resource = hub.tool.boto3.resource.create(ctx, "ec2", "InternetGateway", name)
hub.tool.boto3.resource.exec(resource, attach_to_vpc, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
"""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    tags: List = None,
    vpc_id: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an internet gateway for use with a VPC. After creating the internet gateway, you attach it to a VPC
    using AttachInternetGateway. For more information about your VPC and internet gateway, see the Amazon Virtual
    Private Cloud User Guide.

    Args:
        name(Text): The ID of the internet gateway.
        tags(List, optional): The tags to apply to the resource.
            * Key (string) --  The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.
        vpc_id(List, optional): This list can contain only single element. This is ID of VPC to which internet gateway attaches.
        If vpc_id is not passed, then attach/detach operation is ignored.
        If user passes empty list, then attached vpc is detached.

    Request Syntax:
        [internet_gateway-id]:
          aws.ec2.internet_gateway.present:
          - vpc_id:
            - 'string'
          - tags:
            - Key: 'string'
              Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            igw-54c83a2c:
             aws.ec2.internet_gateway.present:
             - vpc_id:
               - vpc-d04fbf46
             - tags:
               - Key: Name
                 Value: igw-54c83a2c
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "InternetGateway", name)
    # Skip fetching resource if there is no change
    update_new_state = True
    before = await hub.tool.boto3.resource.describe(resource)
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.ec2.internet_gateway {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.ec2.internet_gateway {name}"
            result["result"] = True
        return result

    if before:
        result["old_state"] = before
        result["comment"] = f"'{name}' already exists"
        try:
            if tags is not None:
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=name,
                    old_tags=before.get("Tags", None),
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result
                else:
                    result["comment"] = f"Updated tags '{tags}'"

            ret = await hub.exec.aws.ec2.internet_gateway.update_attachments(
                ctx=ctx,
                internet_gateway_name=name,
                vpc_id=vpc_id,
                before=before,
            )
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            if tags is None and vpc_id is None:
                update_new_state = False
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_internet_gateway(
                ctx,
                TagSpecifications=[{"ResourceType": "internet-gateway", "Tags": tags}]
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            name = ret["ret"]["InternetGateway"]["InternetGatewayId"]
            resource._id = name
            result["comment"] = f"Created '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
        if vpc_id:
            try:
                ret = await hub.exec.boto3.client.ec2.attach_internet_gateway(
                    ctx, InternetGatewayId=name, VpcId=vpc_id[0]
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result[
                    "comment"
                ] = f"Created '{name}' and attached to vpc '{vpc_id[0]}'"
            except hub.tool.boto3.exception.ClientError as e:
                result["result"] = False
                result[
                    "comment"
                ] = f"Created '{name}'. Attach to '{vpc_id[0]}' failed with error: {e}"
    try:
        if update_new_state:
            after = await hub.tool.boto3.resource.describe(resource)
            result["new_state"] = after
        else:
            result["new_state"] = before
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified internet gateway. If resource is attached to VPC. then it is detached first and then deleted.

    Args:
        name(Text): The ID of the internet gateway.

    Request Syntax:
        [internet_gateway-id]:
          aws.ec2.internet_gateway.absent:
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

           igw-54c83a2c:
            aws.ec2.internet_gateway.absent:
              - name: igw-54c83a2c
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "InternetGateway", name)
    before = await hub.tool.boto3.resource.describe(resource)
    result["old_state"] = before
    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.ec2.internet_gateway {name}"
        return result
    else:
        try:
            if before["Attachments"]:
                resource.detach_from_vpc(VpcId=before["Attachments"][0]["VpcId"])
            ret = await hub.exec.boto3.client.ec2.delete_internet_gateway(
                ctx, InternetGatewayId=name
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes one or more of your internet gateways.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.internet_gateway
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_internet_gateways(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe internet_gateway {ret['comment']}")
        return {}

    for internet_gateway in ret["ret"]["InternetGateways"]:
        resource_id = internet_gateway.get("InternetGatewayId")
        resource_translated = []
        if internet_gateway.get("Tags"):
            resource_translated.append({"tags": internet_gateway.get("Tags")})
        if internet_gateway.get("Attachments"):
            resource_translated.append(
                {"vpc_id": [internet_gateway.get("Attachments")[0].get("VpcId")]}
            )
        result[resource_id] = {"aws.ec2.internet_gateway.present": resource_translated}
    return result



"""

Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_security_group
hub.exec.boto3.client.ec2.delete_security_group
hub.exec.boto3.client.ec2.describe_security_groups
hub.tool.boto3.resource.exec(resource, authorize_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, authorize_ingress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_ingress, *args, **kwargs)
"""
from collections import OrderedDict
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.ec2.vpc.present"]},
}


async def present(
    hub,
    ctx,
    name: str,
    description: str,
    vpc_id: str,
    sg_rules: list,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a security group. A security group acts as a virtual firewall for your instance to control inbound and
    outbound traffic. For more information, see Amazon EC2 security groups in the Amazon Elastic Compute Cloud User
    Guide and Security groups for your VPC in the Amazon Virtual Private Cloud User Guide. When you create a
    security group, you specify a friendly name of your choice. You can have a security group for use in EC2-Classic
    with the same name as a security group for use in a VPC. However, you can't have two security groups for use in
    EC2-Classic with the same name or two security groups for use in a VPC with the same name. You have a default
    security group for use in EC2-Classic and a default security group for use in your VPC. If you don't specify a
    security group when you launch an instance, the instance is launched into the appropriate default security
    group. A default security group includes a default rule that grants instances unrestricted network access to
    each other. You can add or remove rules from your security groups using AuthorizeSecurityGroupIngress,
    AuthorizeSecurityGroupEgress, RevokeSecurityGroupIngress, and RevokeSecurityGroupEgress. For more information
    about VPC security group limits, see Amazon VPC Limits.

    Args:
        hub:
        ctx:
        name(str): A name or ID to identify the resource.
        description(str): Description of the security group
        vpc_id(str): Id of the VPC security group should be attached to
        sg_rules(list): Security Group rules to be applied on security group
        tags(Dict, Optional): List of Name, Value pair to tag security group

    Request Syntax:
        [security_group_id]:
          aws.ec2.security_group.present:
          - sg_rules:
            - CidrIpv4: 'string'
              FromPort: 'int'
              IpProtocol: 'string'
              IsEgress: boolean
              Tags:
                - Key: 'string'
                  Value: 'string'
              ToPort: 'int'
          - name: 'string'
          - vpc_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            sg-0008bd25b7867b5cf:
              aws.ec2.security_group.present:
                  - name: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                  - vpc_id: vpc-247e9a5d
                  - description: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                  - sg_rules:
                    - CidrIpv4: 0.0.0.0/0
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    created_sg_id = name
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", name)
    before = await hub.tool.boto3.resource.describe(resource)
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.ec2.security_group {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.ec2.security_group {name}"
            result["result"] = True
        return result
    if before:
        result["old_state"] = before
        update_ret = await hub.exec.aws.ec2.security_group.update_rules_in_sg(
            ctx, created_sg_id, sg_rules, resource
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_security_group(
                ctx,
                GroupName=name,
                Description=description,
                VpcId=vpc_id,
                TagSpecifications=[{"ResourceType": "security-group", "Tags": tags}]
                if tags is not None
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
            created_sg_id = ret["ret"]["GroupId"]
            resource._id = created_sg_id

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

        if sg_rules is not None:
            new_rules_dict = hub.tool.aws.security_group_utils.convert_sg_rules_to_ingress_and_egress(
                sg_rules
            )
            if new_rules_dict["ingress"]:
                new_ingress_rules = hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    new_rules_dict["ingress"]
                )
                [rule.pop("SecurityGroupRuleId", None) for rule in new_ingress_rules]
                ingress_result = (
                    await hub.exec.boto3.client.ec2.authorize_security_group_ingress(
                        ctx,
                        GroupId=created_sg_id,
                        IpPermissions=new_ingress_rules,
                    )
                )
                if not ingress_result["result"]:
                    result["comment"] = ingress_result["comment"]
                    result["result"] = False
            if new_rules_dict["egress"]:
                new_egress_rules = hub.tool.aws.security_group_utils.remove_unwanted_attributes_from_rule(
                    new_rules_dict["egress"]
                )
                [rule.pop("SecurityGroupRuleId", None) for rule in new_egress_rules]
                egress_result = (
                    await hub.exec.boto3.client.ec2.authorize_security_group_egress(
                        ctx,
                        GroupId=created_sg_id,
                        IpPermissions=new_egress_rules,
                    )
                )
                if not egress_result["result"]:
                    result["comment"] = egress_result["comment"]
                    result["result"] = False
    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["new_state"] = after
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a security group. If you attempt to delete a security group that is associated with an instance, or is
    referenced by another security group, the operation fails with InvalidGroup.InUse in EC2-Classic or
    DependencyViolation in EC2-VPC.

    Args:
        hub:
        ctx:
        name(Text): A name, ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.security_group.absent:
                - name: value
    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' is already in deleted state."
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.ec2.security_group {name}"
        return result
    else:
        result["old_state"] = before
        try:
            ret = await hub.exec.boto3.client.ec2.delete_security_group(
                ctx, GroupId=name
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = f"'{name}' is already in deleted state."
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes the specified security groups or all of your security groups. A security group is for use with
    instances either in the EC2-Classic platform or in a specific VPC. For more information, see Amazon EC2 security
    groups in the Amazon Elastic Compute Cloud User Guide and Security groups for your VPC in the Amazon Virtual
    Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.security_group
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_security_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe security_group {ret['comment']}")
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    describe_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "Tags": "tags",
            "Description": "description",
        }
    )
    for security_group in ret["ret"]["SecurityGroups"]:
        translated_security_group = []
        security_group_id = security_group.get("GroupId")
        translated_security_group.append({"name": security_group_id})
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if security_group.get(parameter_old_key) is not None:
                translated_security_group.append(
                    {parameter_new_key: security_group.get(parameter_old_key)}
                )
        describe_sg_rules = (
            await hub.exec.boto3.client.ec2.describe_security_group_rules(
                ctx, Filters=[{"Name": "group-id", "Values": [security_group_id]}]
            )
        )
        if describe_sg_rules and describe_sg_rules["result"]:
            translated_security_group.append(
                {"sg_rules": describe_sg_rules["ret"].get("SecurityGroupRules")}
            )

        result[security_group_id] = {
            "aws.ec2.security_group.present": translated_security_group
        }
    return result
